;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "com-css-sort" "20260101.558"
  "Common way of sorting the CSS attributes."
  '((emacs "25.1")
    (s     "1.12.0"))
  :url "https://github.com/jcs-elpa/com-css-sort"
  :commit "244f910ef55c853e75b0109b7fd96ce742a13461"
  :revdesc "244f910ef55c"
  :keywords '("convenience" "matching" "css" "sort")
  :authors '(("Jen-Chieh" . "jcs090218@gmail.com"))
  :maintainers '(("Jen-Chieh" . "jcs090218@gmail.com")))
